from pathlib import Path
from typing import Tuple, List, Optional
from collections import OrderedDict
from random import randrange

from dovado_rtl.antlr.hdl_representation import ParameterTypeEnum
from dovado_rtl.config import Configuration
import dovado_rtl.vivado_interaction as vivado
from dovado_rtl.frame_handling import (
    TclFrameHandler,
    HdlBoxFrameHandler,
    XdcFrameHandler,
)
from dovado_rtl.enums import RTL, StopStep
from dovado_rtl.simple_types import IsIncremental
from dovado_rtl.point_evaluation import DesignPointEvaluator
from dovado_rtl.fitness import FitnessEvaluator
from dovado_rtl.genetic_algorithm import optimize
from dovado_rtl.cli_utility import (
    validate_board,
    validate_parameters,
    validate_clock_port,
    validate_directives,
    validate_target_clock,
    validate_estimator,
    validate_int_metrics,
    validate_controller,
)
import typer

from dovado_rtl.src_parsing import SourceParser

app = typer.Typer(help="Dovado RTL design automation and space exploration")
vivado.start()


def main():
    app()


@app.callback()
def dovado(
    ctx: typer.Context,
    # Arguments
    # file_path is used through the context in cli_utility during parameters validation
    # and is then passed through context
    # until it reaches this method where it is extracted and used as parsed_src
    file_path: Path = typer.Option(  # pylint: disable=unused-argument
        ...,
        exists=True,
        file_okay=True,
        readable=True,
        resolve_path=True,
        help=(
            "path of the rtl file containing the top level entity, please check the documentation"
            + " for the directory structure your RTL project should comply with"
        ),
    ),
    # Required Cli options
    board: str = typer.Option(
        ...,
        callback=validate_board,
        help="Part name for synthesis/implementation, inserting the wrong one"
        + " will get an error message with all the parts available with your Vivado installation",
    ),
    parameters: List[str] = typer.Option(
        ...,
        callback=validate_parameters,
        help="parameters to use for point/space exploration, only integer (subtypes) and booleans are supported",
    ),
    clock_port: str = typer.Option(
        ...,
        callback=validate_clock_port,
        help="clock port name of the top module/entity",
    ),
    # Optional Cli options
    implementation: bool = typer.Option(
        False,
        help="flag to set point/space exploration to stop at implementation instead of synthesis",
    ),
    incremental: bool = typer.Option(
        False,
        help="flag to choose whether to choose incrementaly synthesis/implementation",
    ),
    directives: Tuple[str, str, str] = typer.Option(
        default=(
            "runtimeoptimized",
            "RuntimeOptimized",
            "RuntimeOptimized",
        ),  # capitalization is necessary to comply with vivado directives specification
        callback=validate_directives,
        help="directives to pass respectively to synthesis, place and route",
    ),
    target_clock: float = typer.Option(
        default=1000,
        callback=validate_target_clock,
        help="target clock (Mhz) on which the worst negative slack is computed,"
        + "make sure this is sufficiently large to never be reached by your design",
    ),
    metrics: Optional[List[int]] = typer.Option(
        None,
        callback=validate_int_metrics,
        help="list of integers representing selected metrics, "
        + "wait for first synthesis/implementation if you do not know the mapping",
    ),
):  # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    config = Configuration()
    Path(str(config.get_config("WORK_DIR"))).mkdir(parents=True, exist_ok=True)

    parsed_src = ctx.obj
    project_path = parsed_src.get_root_folder()

    if implementation:
        stop_step = StopStep.IMPLEMENTATION
    else:
        stop_step = StopStep.SYNTHESIS
    if incremental:
        incremental_mode = IsIncremental(True, True)
    else:
        incremental_mode = IsIncremental(False, False)

    XdcFrameHandler(
        str(config.get_config("PLACEHOLDER")),
        str(config.get_config("XDC_DIR"))
        + str(config.get_config("CONSTRAINT_FRAME")),
        1000 / target_clock,
        str(config.get_config("WORK_DIR"))
        + str(config.get_config("CONSTRAINT")),
    ).fill()

    tcl_handler = TclFrameHandler(
        config,
        parsed_src,
        str(project_path),
        board,
        directives[0],
        incremental_mode,
        stop_step,
        directives[1],
        directives[2],
    )
    tcl_handler.fill()

    box_handler = HdlBoxFrameHandler(
        str(config.get_config("PLACEHOLDER")),
        str(config.get_config("VHDL_DIR"))
        + str(config.get_config("VHDL_BOX_FRAME"))
        if parsed_src.get_hdl() is RTL.VHDL
        else str(config.get_config("VERILOG_DIR"))
        + str(config.get_config("VERILOG_BOX_FRAME")),
        parsed_src.get_selected_entity().get_name(),
        parsed_src.get_ports(),
        parsed_src.get_port(clock_port),
        str(config.get_config("WORK_DIR")) + str(config.get_config("VHDL_BOX"))
        if parsed_src.get_hdl() is RTL.VHDL
        else str(config.get_config("WORK_DIR"))
        + str(config.get_config("VERILOG_BOX")),
        parsed_src.get_hdl(),
        parsed_src.get_folder(),
    )

    point_evaluator = DesignPointEvaluator(
        config,
        parsed_src,
        box_handler,
        tcl_handler,
        target_clock,
        incremental_mode,
        stop_step,
        list(parameters),
        metrics,
    )
    ctx.obj = {
        "parameters": parameters,
        "point_evaluator": point_evaluator,
        "config": config,
    }


@app.command("points")
def points(
    ctx: typer.Context,
    param_values_path: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        readable=True,
        resolve_path=True,
        help="path to a csv file containing on each line values for each parameter previously selected"
        + " (values won't be recorded until the first valid design point, e.g. the first set"
        + " of values which won't make synthesis/implementation fail)",
    ),
):
    """
    RTL Design automation
    """
    out_file = Path(ctx.obj["config"].get_config("WORK_DIR")).joinpath(
        "point_evaluation.csv"
    )
    out_file.open("w").close()
    first_evaluation = True
    for input_line in Path(param_values_path).read_text().splitlines():
        design_point = input_line.split(",")
        design_value = ctx.obj["point_evaluator"].evaluate(tuple(design_point))
        if design_value and first_evaluation:
            output_line = "Design Point,"
            for k in design_value.value.keys():
                if k.utilisation:
                    output_line += k.utilisation[0] + "-" + k.utilisation[1]
                elif k.is_frequency:
                    output_line += "Frequency"
                elif k.custom_metric:
                    output_line += k.custom_metric[0]
                output_line += ","
            out_file.open("a").writelines([output_line[:-1] + "\n"])
            first_evaluation = False
        if design_value and not first_evaluation:
            output_line = ""
            output_line += "("
            for point_value in design_point:
                output_line += point_value + "-"
            output_line = output_line[:-1]
            output_line += "),"
            for v in design_value.value.values():
                output_line += str(v) + ","
            out_file.open("a").writelines([output_line[:-1] + "\n"])


@app.command(
    "space",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def space(
    ctx: typer.Context,
    param_ranges: List[int] = typer.Argument(
        ...,
        help="list of integers in which each odd indexed element starts"
        + " a range and the consecutive closes it (start counting from 1)",
    ),
    power_of_2: Optional[List[str]] = typer.Option(
        None,
        help="list of 'y'/'n' where y indicates that the corresponding, in order, parameter must be a power of 2",
    ),
    param_initial_values: Optional[List[int]] = typer.Option(
        None,
        help="state initial values for parameters which are guaranteed to synthesize/implement",
    ),
    optimization_runtime: Optional[str] = typer.Option(
        None, help="optimization timeout as hh:mm:ss"
    ),
    disable_approximate: Optional[bool] = typer.Option(
        False, help="disable approximations for fitness function"
    ),
    estimation_model: Optional[str] = typer.Option(
        "HoeffdingAdaptiveTree",
        callback=validate_estimator,
        help="estimator to use for fitness function approximation (see movado docs for more information),"
        + "'HoeffdingAdaptiveTree' is default. 'KernelRidge' selects a kernel regression",
    ),
    controller_model: Optional[str] = typer.Option(
        "Mab",
        callback=validate_controller,
        help="controller to use for fitness function approximation (see movado docs for more information),"
        + "'Distance' is default. 'Mab' selects a pure multi-armed bandit controller, ",
    ),
    disable_controller_mab_weight: Optional[bool] = typer.Option(
        False, help="disable loss weighting in Distance Controller"
    ),
    n_controllers: Optional[int] = typer.Option(
        1, help="set the number of controllers for movado (voters)"
    ),
    many_objective: Optional[bool] = typer.Option(
        False,
        help="use AGE-MOEA instead of NSGA-II as genetic algorithm,"
        + " in case of many objectives (e.g. > 4) it may give better performance",
    ),
):
    """
    RTL design space exploration
    """
    for idx, param in enumerate(ctx.obj["parameters"]):
        parsed_src: SourceParser = ctx.obj["point_evaluator"].get_parsed_src()
        full_param = parsed_src.get_parameter(param)
        if full_param.get_type().type is ParameterTypeEnum.BOOL:
            param_ranges = (
                list(param_ranges[: idx * 2])
                + [0, 1]
                + (
                    list(param_ranges[idx * 2 :])
                    if (idx * 2) < len(param_ranges)
                    else []
                )
            )

    it = iter(param_ranges)
    ranges_dict = OrderedDict(zip(ctx.obj["parameters"], zip(it, it)))
    if param_initial_values:
        design_value = ctx.obj["point_evaluator"].evaluate(
            tuple(param_initial_values)
        )
        if not design_value:
            raise Exception(
                "Could not synthesize/implement " + str(param_initial_values)
            )
    else:
        print(
            "Trying random parameter values in the range given to get first synthesis/implementation"
        )
        design_value = ctx.obj["point_evaluator"].evaluate(
            tuple([randrange(i, j) for i, j in ranges_dict.values()])
        )
        while not design_value:
            print(
                "Trying random parameter values in the range given to get first synthesis/implementation"
            )
            design_value = ctx.obj["point_evaluator"].evaluate(
                tuple([randrange(i, j) for i, j in ranges_dict.values()])
            )
        print("First synthesis succeeded")

    metrics = ctx.obj["point_evaluator"].get_metrics()
    fitness_evaluator = FitnessEvaluator(
        ctx.obj["point_evaluator"],
        disable=disable_approximate,
        controller=controller_model,
        estimator=estimation_model,
        controller_mab_weight=not disable_controller_mab_weight,
        voters=n_controllers,
    )

    execution_time = optimize(
        fitness_evaluator,
        ranges_dict,
        metrics,
        optimization_runtime,
        power_of_2,
        many_objective,
    )
    print("\n\nExecution Time: " + str(execution_time) + "\n\n")


main()



from click import BadParameter
import re
import typer
from typing import Tuple, List, Dict, Optional
from pathlib import Path
import os
import importlib
import importlib.util
import importlib.machinery
import sys
from inspect import signature, _empty
import inspect
from dovado_rtl.doc_parsing import get_directives, get_directives_paragraph
from dovado_rtl.src_parsing import SourceParser
from dovado_rtl.vivado_interaction import get_parts
from dovado_rtl.antlr.hdl_representation import PortDirectionEnum
from dovado_rtl.simple_types import Metric
from movado import models, controllers


def validate_power_of_2(value: List[str], ctx: typer.Context):
    if len(ctx.params["parameters"]) != len(value):
        raise BadParameter(
            "power of 2 constraint must be specified for each parameter, mismatching list lengths"
            + ", parameter list lenght:"
            + str(len(ctx.params["parameters"]))
            + ", constraint list length: "
            + str(len(value))
        )
    for i in value:
        if i not in {"y", "n"}:
            raise BadParameter("Accepted input are only y/n")
    return value


def validate_int_metrics(value: List[int]):
    for i in value:
        if i < 0:
            raise BadParameter(
                "Invalid metric '" + str(i) + ", all parameters must be >= 0"
            )
    return value


def parse_comma_separated_list(
    regexp_element: str, to_parse: str,
) -> Optional[List[str]]:
    return (
        None
        if not re.fullmatch(
            "^("
            + regexp_element
            + ")"
            + "(([ \t]*,[ \t]*)("
            + regexp_element
            + "))*$",
            to_parse,
        )
        else [i for i in re.sub(r"[ \t]*", "", to_parse).split(",")]
    )


def __is_mod_function(mod, func):
    "checks that func is a function defined in module mod"
    return inspect.isfunction(func) and inspect.getmodule(func) == mod


def __list_functions(mod):
    "list of functions defined in module mod"
    return [
        func.__name__
        for func in mod.__dict__.values()
        if __is_mod_function(mod, func)
    ]


def __import_path(path: str):
    module_name = os.path.basename(path).replace("-", "_")
    spec = importlib.util.spec_from_loader(
        module_name, importlib.machinery.SourceFileLoader(module_name, path)
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[module_name] = module
    return module


def __is_custom_metric_valid(
    module_function, module_functions: List[str], custom_metric_file: Path
):
    module_function_signature = signature(module_function)
    module_function_parameters = list(
        module_function_signature.parameters.keys()
    )
    module_function_signature_return = (
        module_function_signature.return_annotation
    )

    if len(module_functions) == 0:
        typer.echo(
            "Custom metric file '"
            + str(custom_metric_file)
            + "' contains no public function, custom metric function name must not start with '__'"
        )
        return False
    if len(module_functions) > 1:
        typer.echo(
            "Custom metric file '"
            + str(custom_metric_file)
            + "' contains more than one public function: "
            + str(module_functions)
            + " prepend '__' to all non-public functions"
        )
        return False
    if (
        len(module_function_parameters) > 1
        or len(module_function_parameters) == 0
        or module_function_parameters[0] != "kwargs"
    ):
        typer.echo(
            "Custom metric function '"
            + str(module_functions[0])
            + "' must have **kwargs as parameter instead '"
            + str(module_function_parameters)
            + "' found"
        )
        return False
    if module_function_signature_return != float:
        if module_function_signature_return == _empty:
            typer.echo(
                "WARNING: custom metric function '"
                + str(module_functions[0])
                + "' does not have a return type annotated"
                + " please annotate with -> float, the only return value supported"
            )
        else:
            typer.echo(
                "Custom metric function '"
                + str(module_functions[0])
                + "' must have 'float' as a return type + '"
                + str(module_function_signature)
                + "' found instead"
            )
            return False
    return True


def ask_utilization_metrics(
    util_indices: Dict[str, List[str]],
    custom_metrics_path: Path,
    pre_selected_metrics: Optional[List[int]] = None,
) -> List[Metric]:
    info_prompt = "Percentage Utilisation indices available:\n"
    i = 0
    counter_dict = {}
    for section in util_indices.keys():
        info_prompt += section + "\n"
        for util_metric in util_indices[section]:
            i = i + 1
            counter_dict[i] = Metric(
                utilisation=(section, util_metric), is_frequency=False
            )
            info_prompt += "\t(" + str(i) + ") " + util_metric + "\n"
    if not custom_metrics_path.is_dir():
        typer.echo(
            "No custom metrics folder found at: " + str(custom_metrics_path)
        )
    else:
        custom_metric_dict = {}
        for custom_metric_file in custom_metrics_path.iterdir():
            if custom_metric_file.is_dir():
                continue
            module = __import_path(str(custom_metric_file))
            module_functions = [
                function
                for function in __list_functions(module)
                if function[0] != "_" and function[1] != "_"
            ]
            module_function = getattr(module, module_functions[0])
            if __is_custom_metric_valid(
                module_function, module_functions, custom_metric_file
            ):
                custom_metric_dict[module_functions[0]] = module_function
        info_prompt += "Custom Metrics Available:\n"
        for (
            custom_metric_name,
            custom_metric_function,
        ) in custom_metric_dict.items():
            i = i + 1
            counter_dict[i] = Metric(
                utilisation=None,
                is_frequency=False,
                custom_metric=(custom_metric_name, custom_metric_function),
            )
            info_prompt += "\t(" + str(i) + ") " + custom_metric_name + "\n"

    while True:
        if not pre_selected_metrics:
            typer.echo(info_prompt)
            user_input = typer.prompt(
                "Enter a comma-separated list of percentage "
                + "indices to optimize, input 0 to select frequency"
            )
        try:
            if not pre_selected_metrics:
                parsed_list = [
                    int(i)
                    for i in parse_comma_separated_list(r"[+]?\d+", user_input)
                ]
            else:
                parsed_list = pre_selected_metrics
                pre_selected_metrics = None
            if parsed_list and (
                parsed_list == []
                or any(
                    [i in counter_dict.keys() for i in parsed_list if i > 0]
                )
            ):
                out: List[Metric] = [
                    counter_dict[i] for i in {i for i in parsed_list if i > 0}
                ]
                return (
                    (out + [Metric(None, True)]) if (0 in parsed_list) else out
                )
        except Exception as e:
            typer.echo(str(e))
            typer.echo(
                "Invalid input, please enter"
                + " a comma separated list of integer numbers"
            )


def validate_nearest_distance(value: int) -> int:
    if value < 0:
        raise typer.BadParameter(
            "Invalid distance value: "
            + str(value)
            + " non-negative integer expected"
        )
    return value


def validate_estimator(value: str) -> str:
    if value in models:
        return value
    raise typer.BadParameter(
        "Invalid estimation model "
        + value
        + " available models are: "
        + str(models)
    )


def validate_controller(value: str) -> str:
    if value in controllers:
        return value
    raise typer.BadParameter(
        "Invalid controller "
        + value
        + " available controllers are: "
        + str(controllers)
    )


def validate_board(value: str) -> str:
    parts = get_parts()
    if value in set(parts.keys()):
        return value
    raise typer.BadParameter(
        "Invalid board " + value + " available boards are: " + str(parts)
    )


def validate_parameters(ctx: typer.Context, value: List[str]) -> List[str]:

    file_path = Path(ctx.params["file_path"]).absolute()
    if not ctx.obj:
        parsed_src = SourceParser(
            str(SourceParser.get_project_path(file_path)), str(file_path)
        )
        parsed_src.set_entity(str(parsed_src.get_entities()[0].get_name()))
        ctx.obj = parsed_src
    else:
        parsed_src: SourceParser = ctx.obj
    parameter_name_list = [p.get_name() for p in parsed_src.get_parameters()]
    for chosen_parameter in value:
        if chosen_parameter not in parameter_name_list:
            raise typer.BadParameter(
                "Chosen parameter: "
                + chosen_parameter
                + " is not among available parameters "
                + str(parameter_name_list)
            )
    return value


def validate_clock_port(ctx: typer.Context, value: str) -> str:
    parsed_src: SourceParser = ctx.obj
    for port in parsed_src.get_ports():
        if (
            port.get_name() == value
            and port.get_direction().direction is PortDirectionEnum.INPUT
        ):
            return value
    raise typer.BadParameter(
        "Chosen clock port: "
        + value
        + " is not among available input ports "
        + str(
            [
                p.get_name()
                for p in parsed_src.get_ports()
                if p.get_direction().direction is PortDirectionEnum.INPUT
            ]
        )
    )


def validate_directives(value: Tuple[str, str, str]):
    commands = ["synthesis", "place", "route"]
    for c in commands:
        directive = value[commands.index(c)]
        directives = get_directives((get_directives_paragraph(c)))
        if directive not in directives:
            raise typer.BadParameter(
                "Chosen directive for "
                + c
                + ": "
                + directive
                + " is not among available directives for "
                + c
                + " "
                + str(directives)
            )
    return value


def validate_target_clock(value: float) -> float:
    if value > 0:
        return value
    raise BadParameter(
        "Target clock must be positive, invalid target clock input: "
        + str(value)
    )


def validate_param_ranges(value: List[int]) -> List[int]:
    if len(value) % 2 == 0:
        return value
    raise typer.BadParameter(
        "Parameter ranges list must be of even length, "
        + "each odd element is the start of the range and the next one is the end of the range"
    )



from pathlib import Path
from typing import List, Optional
from dovado_rtl.src_parsing import SourceParser
from dovado_rtl.config import Configuration
from dovado_rtl.antlr.hdl_representation import (
    Parameter,
    Port,
    PortDirectionEnum,
    PortTypeEnum,
)
from dovado_rtl.enums import RTL, StopStep
from dovado_rtl.simple_types import IsIncremental
from dovado_rtl.fill_handler import FillHandler


class XdcFrameHandler(FillHandler):
    def __init__(
        self,
        placeholder: str,
        frame_path: str,
        target_clock: float,
        out_path: str,
    ):
        self.frame_path: str = frame_path
        self.target_clock: float = target_clock
        self.out_path: str = out_path
        FillHandler.__init__(self, placeholder, frame_path, out_path)

    def fill(self) -> bool:
        self.replacements = [str(self.target_clock)]
        return super().fill()


class TclFrameHandler(FillHandler):
    def __init__(
        self,
        config: Configuration,
        parsed_source: SourceParser,
        base_folder: str,
        synthesis_part: str,
        synthesis_directive: str,
        incremental_mode: IsIncremental,
        stop_step: StopStep,
        place_directive: Optional[str] = None,
        route_directive: Optional[str] = None,
    ):
        self.parsed_source: SourceParser = parsed_source
        self.base_folder: str = base_folder
        self.synthesis_part: str = synthesis_part
        self.synthesis_directive: str = synthesis_directive
        self.incremental_mode: IsIncremental = incremental_mode
        self.stop_step: StopStep = stop_step
        self.place_directive: Optional[str] = place_directive
        self.route_directive: Optional[str] = route_directive
        self.config: Configuration = config

        FillHandler.__init__(
            self,
            str(config.get_config("PLACEHOLDER")),
            str(self.config.get_config("TCL_DIR"))
            + str(self.config.get_config(self.stop_step.name + "_FRAME")),
            str(self.config.get_config("WORK_DIR"))
            + str(self.config.get_config(self.stop_step.name)),
        )

    def fill(self) -> bool:
        top_lang: RTL = self.parsed_source.get_hdl()
        is_library = self.parsed_source.is_library_project()
        user_libs = "{"
        folder = ""
        if is_library:
            for lib in self.parsed_source.get_user_defined_libs():
                user_libs += " " + lib
        else:
            folder = self.parsed_source.get_folder()
            user_libs += " " + folder
        user_libs += " }"

        synthesis_replacements: List[str] = [
            str(self.config.get_config("WORK_DIR")),
            self.base_folder,
            str(self.config.get_config("WORK_DIR"))
            + str(self.config.get_config("CONSTRAINT")),
            user_libs,
            (
                "read_vhdl -library "
                + ("dovado " if is_library else (folder + " "))
                + str(self.config.get_config("WORK_DIR"))
                + str(self.config.get_config("VHDL_BOX"))
            )
            if top_lang is RTL.VHDL
            else (
                "read_verilog -sv -library "
                + ("dovado " if is_library else (folder + " "))
                + str(self.config.get_config("WORK_DIR"))
                + str(self.config.get_config("VERILOG_BOX"))
            ),
            "box",
            self.synthesis_part,
            self.synthesis_directive,
            "-incremental_synth" if self.incremental_mode.synthesis else "",
        ]

        implementation_replacements: Optional[List[str]] = (
            []
            if self.stop_step is StopStep.SYNTHESIS
            else (
                [
                    "",
                    "-directive " + str(self.place_directive),
                    "",
                    "-directive " + str(self.route_directive),
                ]
                if not self.incremental_mode.implementation
                else [
                    "-directive " + str(self.place_directive),
                    " ",
                    "-directive " + str(self.route_directive),
                    " ",
                ]
            )
        )

        self.replacements = (
            synthesis_replacements
            if self.stop_step is StopStep.SYNTHESIS
            else synthesis_replacements[:4]
            + [
                (
                    "read_vhdl -library "
                    + ("dovado " if is_library else (folder + " "))
                    + str(self.config.get_config("WORK_DIR"))
                    + str(self.config.get_config("VHDL_BOX"))
                )
                if top_lang is RTL.VHDL
                else (
                    "read_verilog -sv -library "
                    + ("dovado " if is_library else (folder + " "))
                    + str(self.config.get_config("WORK_DIR"))
                    + str(self.config.get_config("VERILOG_BOX"))
                )
            ]
            + synthesis_replacements[5:]
            + implementation_replacements
        )
        self.replacements += [
            str(self.config.get_config(self.stop_step.name + "_TIMING")),
            str(self.config.get_config(self.stop_step.name + "_UTILISATION")),
        ]
        return super().fill()

    def setup_incremental(self, is_incremental: IsIncremental):
        if not is_incremental.synthesis and not is_incremental.implementation:
            return
        path = str(self.config.get_config("WORK_DIR"))
        if is_incremental.synthesis:
            path += str(self.config.get_config("SYNTHESIS"))
        if is_incremental.implementation:
            path += str(self.config.get_config("IMPLEMENTATION"))
        text = Path(path).read_text()
        text = text.replace("#! ", "")
        Path(path).write_text(text)


class HdlBoxFrameHandler(FillHandler):
    def __init__(
        self,
        placeholder: str,
        frame_path: str,
        top_module: str,
        ports: List[Port],
        clock_port: Port,
        out_path: str,
        hdl: RTL,
        lib: Optional[str],
    ):
        self.__top_module: str = top_module
        self.__ports: List[Port] = ports
        self.__clock_port: Port = clock_port
        self.__parameters: Optional[List[Parameter]] = None
        self.__hdl: RTL = hdl
        self.__lib: Optional[str] = lib

        FillHandler.__init__(self, placeholder, frame_path, out_path)

    def get_parameters(self) -> List[Parameter]:
        if self.__parameters:
            return self.__parameters
        raise Exception(
            "Trying to access parameters in HdlBoxFrameHandler before setting them"
        )

    def set_parameters(self, parameters: List[Parameter]) -> None:
        self.__parameters = parameters

    def fill(self) -> bool:
        if not self.__parameters:
            raise Exception(
                "Cannot fill box without setting parameters beforehand"
            )

        if self.__hdl is RTL.VHDL:
            if self.__lib:
                return self.__vhdl_fill_box(self.__lib)
            raise ValueError(
                "Cannot retrieve lib in which the top level resides"
            )
        return self.__verilog_fill_box()

    def __vhdl_parameter_map(self) -> str:
        parameter_section: str = "generic map(\n"
        for parameter in self.get_parameters()[:-1]:
            if parameter.get_value() is None:
                raise Exception(
                    "Please provide default values for all parameters, "
                    + str(parameter.get_name())
                    + " does not have one."
                    + "\n Parameter List:\n"
                    + str(self.__parameters)
                )
            parameter_section += (
                parameter.get_name()
                + " => "
                + (str(parameter.get_value())).strip()
                + ",\n"
            )
        parameter_section += (
            self.get_parameters()[-1].get_name()
            + " => "
            + (str(self.get_parameters()[-1].get_value())).strip()
            + ")"
        )
        return parameter_section

    def __vhdl_fill_box(self, lib: str) -> bool:
        input_mapping = [
            port.get_name()
            + (
                " => " + port.get_type().descriptor + "'('1'),\n"
                if port.get_type().type is PortTypeEnum.SCALAR
                else " => "
                + port.get_type().descriptor
                + "'((others => std_logic'('1'))),\n"
            )
            for port in self.__ports
            if (
                port.get_direction().direction is PortDirectionEnum.INPUT
                and port.get_name() != self.__clock_port.get_name()
            )
        ]
        input_mapping[len(input_mapping) - 1] = input_mapping[
            len(input_mapping) - 1
        ].replace(",", "")
        self.replacements = [
            "library " + lib + ";",
            lib + "." + self.__top_module,
            self.__vhdl_parameter_map(),
            self.__clock_port.get_name(),
            "".join(input_mapping),
        ]
        return super().fill()

    def __verilog_parameter_map(self) -> str:
        parameter_section: str = "#(\n"
        for parameter in self.get_parameters()[:-1]:
            parameter_section += (
                "."
                + parameter.get_name()
                + "("
                + str(parameter.get_value())
                + "),\n"
            )
        parameter_section += (
            "."
            + self.get_parameters()[-1].get_name()
            + "("
            + str(self.get_parameters()[-1].get_value())
            + ")\n)\n"
        )
        return parameter_section

    def __verilog_fill_box(self) -> bool:
        input_mapping = [
            "." + port.get_name() + " ('1),\n"
            for port in self.__ports
            if (
                port.get_direction().direction is PortDirectionEnum.INPUT
                and port.get_name() != self.__clock_port.get_name()
            )
        ]

        input_mapping[len(input_mapping) - 1] = input_mapping[
            len(input_mapping) - 1
        ].replace(",", "")

        self.replacements = [
            self.__top_module,
            self.__verilog_parameter_map(),
            self.__clock_port.get_name(),
            "".join(input_mapping),
        ]
        return super().fill()
